#!/usr/bin/python
# -*- coding: UTF-8 -*-

from _collections import defaultdict
from datetime import datetime
from pytz import timezone

import json

from .extra import DatabaseExtended as _DatabaseExtended

################################################################################
# 
################################################################################

defDict = lambda: defaultdict( defDict )

################################################################################
# 
################################################################################

try:
    from NIST.fingerprint import NISTf
    from NIST.fingerprint.functions import lstTo012

except:
    raise Exception( "Can not import the NIST module. Check the installation." )

else:
    
    minutype_NIST2PiAnoS = {
        'A': "ridge_ending",
        'B': "bifurcation",
        'C': "unknown",
        'D': "unknown"
    }
        
    minutype_PiAnoS2NIST = {
        "ridge_ending": 'A',
        "bifurcation": 'B',
        "unknown": 'D'
    }
        
    class DatabaseExtended( _DatabaseExtended ):
        ########################################################################
        # 
        #    Create case in database based on a NIST object.
        # 
        ########################################################################
        
        def create_exercise_NIST( self, **kwargs ):
            """
                Import the NIST object and create the exercise.
            """
            folder_id = kwargs.get( "folder_id" )
            name = kwargs.get( "name" )
            description = kwargs.get( "description" )
            mark = kwargs.get( "mark" )
            pr = kwargs.get( "pr" )
            markidc = kwargs.get( "markidc", -1 )
            printidc = kwargs.get( "pridc", -1 )
            
            markidc = int( markidc )
            printidc = int( printidc )
            
            return self.create_exercise( 
                folder_id = folder_id,
                name = name,
                description = description,
                mark = mark.get_latent( 'PIL', markidc ),
                mark_resolution = mark.get_resolution( markidc ),
                _print = pr.get_print( 'PIL', printidc ),
                _print_resolution = pr.get_resolution( printidc ),
                markname = mark.get_name(),
                _printname = "%s-%02d" % ( pr.get_name(), printidc )
            )
        
        def update_latent_with_annotations( self, exercise_id, username, minutiae ):
            """
                Update the annotations on a mark present in the database, for
                the case 'exercise_id'. The annotations are made as if the user
                'username' would have made them.
            """
            if type( username ) == str:
                userid = self.get_userID( username )
            else:
                userid = username
 
            sql = "INSERT INTO object ( creator ) VALUES ( '%s' ) RETURNING id" % userid
            objid = self.query( sql ).fetchone()[ 0 ]
             
            sql = "INSERT INTO assignment ( id, exercise, ispreview ) VALUES ( %s, %s, FALSE ) RETURNING id" % ( objid, exercise_id )
            assid = self.query( sql ).fetchone()[ 0 ]
             
            date = timezone( 'Europe/Zurich' ).localize( datetime.now() )
            date = str( date )
            
            d = defDict()
            
            markid, _ = self.get_images_exercise( exercise_id )
            width, height, res = self.get_image_info( markid )
            
            for id, x, y, theta, designation in minutiae:
                id, x, y, theta, designation = NIST2PiAnoS( id, x, y, theta, designation, res, height ) 
                
                ids = str( id )
                
                d[ 'mark' ][ 'minutiae' ][ ids ][ "id" ] = id
                d[ 'mark' ][ 'minutiae' ][ ids ][ "type" ] = designation
                d[ 'mark' ][ 'minutiae' ][ ids ][ "pos" ] = [ x, y ]
                d[ 'mark' ][ 'minutiae' ][ ids ][ "angle" ] = theta
                d[ 'mark' ][ 'minutiae' ][ ids ][ "userid" ] = userid
                 
            data = json.loads( json.dumps( d ) )
             
            data = str( data )
            data = data.replace( "u'", "'" ).replace( "'", '"' )
             
            sql = "INSERT INTO result ( assignment, type, step, updatedon, data ) VALUES ( %s, %s, %s, '%s', '%s' ) RETURNING assignment;" % ( assid, 1, 1, date, data )
            self.query( sql )

        ########################################################################
        # 
        #    Exporting in the NIST format
        # 
        ########################################################################
        
        def export_latent( self, exercisename = None, username = None ):
            n = NISTf()
            n.add_Type01()
            n.add_Type02()
            n.set_field( "2.007", exercisename )
            
            idc = 1
            for f in self.exportExerciseData( exercisename, username ).fetchall():
                data = f[ 'data' ]
                markid = f[ 'mark' ]
            
                width, height, res = self.getImageInformations( markid )
                
                img = self.getImage( markid )
                
                data = json.loads( data )
                
                ret = []
                i = 1
                for key, value in data[ 'mark' ][ 'minutiae' ].iteritems():
                    try:
                        id, x, y, theta, q, type = PiAnoS2NIST( i, value, res, height )
                        ret.append( ( id, x, y, theta, q, type ) )
                        
                        i += 1
                    
                    except KeyError:
                        continue
                
                n.add_Type09( lstTo012( ret ), idc )
                
                n.add_Type13( ( width, height ), res, idc )
                n.set_latent( img, res, idc )
                
                n.set_field( "13.925", f[ 'username' ], idc )
                
                idc += 1
            
            return n
        
        def export_print( self, exercisename = None, username = None, limit = 1 ):
            n = NISTf()
            n.add_Type01()
            n.add_Type02()
            
            f = self.exportExerciseData( exercisename, username, limit ).fetchone()
            data = f[ 'data' ]
            printid = f[ 'print' ]
        
            width, height, res = self.getImageInformations( printid )
            
            data = json.loads( data )
            
            ret = []
            i = 0
            for key, value in data[ 'print' ][ 'minutiae' ].iteritems():
                if value[ 'type' ] == "center":
                    continue
                else:
                    try:
                        id, x, y, theta, q, type = PiAnoS2NIST( i, value, res, height )
                        ret.append( ( id, x, y, theta, q, type ) )
                        
                        i += 1
                    
                    except KeyError:
                        continue
                    
            n.add_Type09( lstTo012( ret ), 1 )
            
            return n
        
    ############################################################################
    # 
    #    Coordinates changes
    # 
    ############################################################################
    
    def NIST2PiAnoS( id, x, y, theta, type, res, height ):
        """
            Change the coordinate system from NIST to PiAnoS
        """
        id = int( id )
                
        x = x / 25.4 * res 
        y = y / 25.4 * res
        
        y = height - y
        
        theta = ( -theta + 180 ) % 360
         
        type = minutype_NIST2PiAnoS[ type ]
        
        return ( id, x, y, theta, type )
    
    def PiAnoS2NIST( id, value, res, height ):
        """
            Change the coordinate system from PiAnoS to NIST
        """
        id = "%03d" % id
        
        x, y = value[ 'pos' ]
        theta = value[ 'angle' ]
        
        x = x / res * 25.4
        y = ( height - y ) / res * 25.4 
        
        theta = ( -theta + 180 ) % 360
        type = minutype_PiAnoS2NIST[ value[ 'type' ] ]
        
        return ( id, x, y, theta, '00', type )
